@@ -60,13 +60,10 @@ MAX_NAME_LENGTH = 253
 
 DECODE_EXCEPTIONS = (IndexError, struct.error, IncomingDecodeError)
 
-UNPACK_3H = struct.Struct(b'!3H').unpack_from
-UNPACK_6H = struct.Struct(b'!6H').unpack_from
-UNPACK_HH = struct.Struct(b'!HH').unpack_from
-UNPACK_HHiH = struct.Struct(b'!HHiH').unpack_from
 
 _seen_logs: Dict[str, Union[int, tuple]] = {}
 _str = str
+_int = int
 
 
 class DNSIncoming:
@@ -77,19 +74,21 @@ class DNSIncoming:
         'flags',
         'offset',
         'data',
+        'view',
         '_data_len',
-        'name_cache',
-        'questions',
+        '_name_cache',
+        '_questions',
         '_answers',
         'id',
-        'num_questions',
-        'num_answers',
-        'num_authorities',
-        'num_additionals',
+        '_num_questions',
+        '_num_answers',
+        '_num_authorities',
+        '_num_additionals',
         'valid',
         'now',
         'scope_id',
         'source',
+        '_has_qu_question',
     )
 
     def __init__(
@@ -103,20 +102,22 @@ class DNSIncoming:
         self.flags = 0
         self.offset = 0
         self.data = data
+        self.view = data
         self._data_len = len(data)
-        self.name_cache: Dict[int, List[str]] = {}
-        self.questions: List[DNSQuestion] = []
+        self._name_cache: Dict[int, List[str]] = {}
+        self._questions: List[DNSQuestion] = []
         self._answers: List[DNSRecord] = []
         self.id = 0
-        self.num_questions = 0
-        self.num_answers = 0
-        self.num_authorities = 0
-        self.num_additionals = 0
+        self._num_questions = 0
+        self._num_answers = 0
+        self._num_authorities = 0
+        self._num_additionals = 0
         self.valid = False
         self._did_read_others = False
         self.now = now or current_time_millis()
         self.source = source
         self.scope_id = scope_id
+        self._has_qu_question = False
         try:
             self._initial_parse()
         except DECODE_EXCEPTIONS:
@@ -135,16 +136,45 @@ class DNSIncoming:
         """Returns true if this is a response."""
         return (self.flags & _FLAGS_QR_MASK) == _FLAGS_QR_RESPONSE
 
+    def has_qu_question(self) -> bool:
+        """Returns true if any question is a QU question."""
+        return self._has_qu_question
+
     @property
     def truncated(self) -> bool:
         """Returns true if this is a truncated."""
         return (self.flags & _FLAGS_TC) == _FLAGS_TC
 
+    @property
+    def questions(self) -> List[DNSQuestion]:
+        """Questions in the packet."""
+        return self._questions
+
+    @property
+    def num_questions(self) -> int:
+        """Number of questions in the packet."""
+        return self._num_questions
+
+    @property
+    def num_answers(self) -> int:
+        """Number of answers in the packet."""
+        return self._num_answers
+
+    @property
+    def num_authorities(self) -> int:
+        """Number of authorities in the packet."""
+        return self._num_authorities
+
+    @property
+    def num_additionals(self) -> int:
+        """Number of additionals in the packet."""
+        return self._num_additionals
+
     def _initial_parse(self) -> None:
         """Parse the data needed to initalize the packet object."""
         self._read_header()
         self._read_questions()
-        if not self.num_questions:
+        if not self._num_questions:
             self._read_others()
         self.valid = True
 
@@ -159,7 +189,6 @@ class DNSIncoming:
             log_exc_info = True
         log.debug(*(logger_data or ['Exception occurred']), exc_info=log_exc_info)
 
-    @property
     def answers(self) -> List[DNSRecord]:
         """Answers in the packet."""
         if not self._did_read_others:
@@ -174,10 +203,9 @@ class DNSIncoming:
                 )
         return self._answers
 
-    @property
     def is_probe(self) -> bool:
         """Returns true if this is a probe."""
-        return self.num_authorities > 0
+        return self._num_authorities > 0
 
     def __repr__(self) -> str:
         return '<DNSIncoming:{%s}>' % ', '.join(
@@ -185,43 +213,53 @@ class DNSIncoming:
                 'id=%s' % self.id,
                 'flags=%s' % self.flags,
                 'truncated=%s' % self.truncated,
-                'n_q=%s' % self.num_questions,
-                'n_ans=%s' % self.num_answers,
-                'n_auth=%s' % self.num_authorities,
-                'n_add=%s' % self.num_additionals,
-                'questions=%s' % self.questions,
-                'answers=%s' % self.answers,
+                'n_q=%s' % self._num_questions,
+                'n_ans=%s' % self._num_answers,
+                'n_auth=%s' % self._num_authorities,
+                'n_add=%s' % self._num_additionals,
+                'questions=%s' % self._questions,
+                'answers=%s' % self.answers(),
             ]
         )
 
     def _read_header(self) -> None:
         """Reads header portion of packet"""
-        (
-            self.id,
-            self.flags,
-            self.num_questions,
-            self.num_answers,
-            self.num_authorities,
-            self.num_additionals,
-        ) = UNPACK_6H(self.data)
+        view = self.view
+        offset = self.offset
         self.offset += 12
+        # The header has 6 unsigned shorts in network order
+        self.id = view[offset] << 8 | view[offset + 1]
+        self.flags = view[offset + 2] << 8 | view[offset + 3]
+        self._num_questions = view[offset + 4] << 8 | view[offset + 5]
+        self._num_answers = view[offset + 6] << 8 | view[offset + 7]
+        self._num_authorities = view[offset + 8] << 8 | view[offset + 9]
+        self._num_additionals = view[offset + 10] << 8 | view[offset + 11]
 
     def _read_questions(self) -> None:
         """Reads questions section of packet"""
-        for _ in range(self.num_questions):
+        view = self.view
+        questions = self._questions
+        for _ in range(self._num_questions):
             name = self._read_name()
-            type_, class_ = UNPACK_HH(self.data, self.offset)
+            offset = self.offset
             self.offset += 4
+            # The question has 2 unsigned shorts in network order
+            type_ = view[offset] << 8 | view[offset + 1]
+            class_ = view[offset + 2] << 8 | view[offset + 3]
             question = DNSQuestion(name, type_, class_)
-            self.questions.append(question)
+            if question.unique:  # QU questions use the same bit as unique
+                self._has_qu_question = True
+            questions.append(question)
 
-    def _read_character_string(self) -> bytes:
+    def _read_character_string(self) -> str:
         """Reads a character string from the packet"""
-        length = self.data[self.offset]
+        length = self.view[self.offset]
         self.offset += 1
-        return self._read_string(length)
+        info = self.data[self.offset : self.offset + length].decode('utf-8', 'replace')
+        self.offset += length
+        return info
 
-    def _read_string(self, length: int) -> bytes:
+    def _read_string(self, length: _int) -> bytes:
         """Reads a string of a given length from the packet"""
         info = self.data[self.offset : self.offset + length]
         self.offset += length
@@ -231,11 +269,18 @@ class DNSIncoming:
         """Reads the answers, authorities and additionals section of the
         packet"""
         self._did_read_others = True
-        n = self.num_answers + self.num_authorities + self.num_additionals
+        view = self.view
+        n = self._num_answers + self._num_authorities + self._num_additionals
         for _ in range(n):
             domain = self._read_name()
-            type_, class_, ttl, length = UNPACK_HHiH(self.data, self.offset)
+            offset = self.offset
             self.offset += 10
+            # type_, class_ and length are unsigned shorts in network order
+            # ttl is an unsigned long in network order https://www.rfc-editor.org/errata/eid2130
+            type_ = view[offset] << 8 | view[offset + 1]
+            class_ = view[offset + 2] << 8 | view[offset + 3]
+            ttl = view[offset + 4] << 24 | view[offset + 5] << 16 | view[offset + 6] << 8 | view[offset + 7]
+            length = view[offset + 8] << 8 | view[offset + 9]
             end = self.offset + length
             rec = None
             try:
@@ -257,20 +302,23 @@ class DNSIncoming:
                 self._answers.append(rec)
 
     def _read_record(
-        self, domain: _str, type_: int, class_: int, ttl: int, length: int
+        self, domain: _str, type_: _int, class_: _int, ttl: _int, length: _int
     ) -> Optional[DNSRecord]:
         """Read known records types and skip unknown ones."""
         if type_ == _TYPE_A:
-            dns_address = DNSAddress(domain, type_, class_, ttl, self._read_string(4))
-            dns_address.created = self.now
-            return dns_address
+            return DNSAddress(domain, type_, class_, ttl, self._read_string(4), None, self.now)
         if type_ in (_TYPE_CNAME, _TYPE_PTR):
             return DNSPointer(domain, type_, class_, ttl, self._read_name(), self.now)
         if type_ == _TYPE_TXT:
             return DNSText(domain, type_, class_, ttl, self._read_string(length), self.now)
         if type_ == _TYPE_SRV:
-            priority, weight, port = UNPACK_3H(self.data, self.offset)
+            view = self.view
+            offset = self.offset
             self.offset += 6
+            # The SRV record has 3 unsigned shorts in network order
+            priority = view[offset] << 8 | view[offset + 1]
+            weight = view[offset + 2] << 8 | view[offset + 3]
+            port = view[offset + 4] << 8 | view[offset + 5]
             return DNSService(
                 domain,
                 type_,
@@ -288,15 +336,12 @@ class DNSIncoming:
                 type_,
                 class_,
                 ttl,
-                self._read_character_string().decode('utf-8', 'replace'),
-                self._read_character_string().decode('utf-8', 'replace'),
+                self._read_character_string(),
+                self._read_character_string(),
                 self.now,
             )
         if type_ == _TYPE_AAAA:
-            dns_address = DNSAddress(domain, type_, class_, ttl, self._read_string(16))
-            dns_address.created = self.now
-            dns_address.scope_id = self.scope_id
-            return dns_address
+            return DNSAddress(domain, type_, class_, ttl, self._read_string(16), self.scope_id, self.now)
         if type_ == _TYPE_NSEC:
             name_start = self.offset
             return DNSNsec(
@@ -314,15 +359,16 @@ class DNSIncoming:
         self.offset += length
         return None
 
-    def _read_bitmap(self, end: int) -> List[int]:
+    def _read_bitmap(self, end: _int) -> List[int]:
         """Reads an NSEC bitmap from the packet."""
         rdtypes = []
+        view = self.view
         while self.offset < end:
             offset = self.offset
             offset_plus_one = offset + 1
             offset_plus_two = offset + 2
-            window = self.data[offset]
-            bitmap_length = self.data[offset_plus_one]
+            window = view[offset]
+            bitmap_length = view[offset_plus_one]
             bitmap_end = offset_plus_two + bitmap_length
             for i, byte in enumerate(self.data[offset_plus_two:bitmap_end]):
                 for bit in range(0, 8):
@@ -337,7 +383,7 @@ class DNSIncoming:
         seen_pointers: Set[int] = set()
         original_offset = self.offset
         self.offset = self._decode_labels_at_offset(original_offset, labels, seen_pointers)
-        self.name_cache[original_offset] = labels
+        self._name_cache[original_offset] = labels
         name = ".".join(labels) + "."
         if len(name) > MAX_NAME_LENGTH:
             raise IncomingDecodeError(
@@ -345,10 +391,11 @@ class DNSIncoming:
             )
         return name
 
-    def _decode_labels_at_offset(self, off: int, labels: List[str], seen_pointers: Set[int]) -> int:
+    def _decode_labels_at_offset(self, off: _int, labels: List[str], seen_pointers: Set[int]) -> int:
         # This is a tight loop that is called frequently, small optimizations can make a difference.
+        view = self.view
         while off < self._data_len:
-            length = self.data[off]
+            length = view[off]
             if length == 0:
                 return off + DNS_COMPRESSION_HEADER_LEN
 
@@ -364,9 +411,9 @@ class DNSIncoming:
                 )
 
             # We have a DNS compression pointer
-            link_data = self.data[off + 1]
+            link_data = view[off + 1]
             link = (length & 0x3F) * 256 + link_data
-            lint_int = int(link)
+            link_py_int = link
             if link > self._data_len:
                 raise IncomingDecodeError(
                     f"DNS compression pointer at {off} points to {link} beyond packet from {self.source}"
@@ -375,16 +422,16 @@ class DNSIncoming:
                 raise IncomingDecodeError(
                     f"DNS compression pointer at {off} points to itself from {self.source}"
                 )
-            if lint_int in seen_pointers:
+            if link_py_int in seen_pointers:
                 raise IncomingDecodeError(
                     f"DNS compression pointer at {off} was seen again from {self.source}"
                 )
-            linked_labels = self.name_cache.get(lint_int)
+            linked_labels = self._name_cache.get(link_py_int)
             if not linked_labels:
                 linked_labels = []
-                seen_pointers.add(lint_int)
+                seen_pointers.add(link_py_int)
                 self._decode_labels_at_offset(link, linked_labels, seen_pointers)
-                self.name_cache[lint_int] = linked_labels
+                self._name_cache[link_py_int] = linked_labels
             labels.extend(linked_labels)
             if len(labels) > MAX_DNS_LABELS:
                 raise IncomingDecodeError(
