@@ -28,7 +28,7 @@ from ._core import Zeroconf
 from ._dns import DNSQuestionType
 from ._services import ServiceListener
 from ._services.browser import _ServiceBrowserBase
-from ._services.info import ServiceInfo
+from ._services.info import AsyncServiceInfo, ServiceInfo
 from ._services.types import ZeroconfServiceTypes
 from ._utils.net import InterfaceChoice, InterfacesType, IPVersion
 from .const import _BROWSER_TIME, _MDNS_PORT, _SERVICE_TYPE_ENUMERATION_NAME
@@ -41,10 +41,6 @@ __all__ = [
 ]
 
 
-class AsyncServiceInfo(ServiceInfo):
-    """An async version of ServiceInfo."""
-
-
 class AsyncServiceBrowser(_ServiceBrowserBase):
     """Used to browse for a service for specific type(s).
 
@@ -82,16 +78,17 @@ class AsyncServiceBrowser(_ServiceBrowserBase):
         """Cancel the browser."""
         self._async_cancel()
 
-    def async_update_records_complete(self) -> None:
-        """Called when a record update has completed for all handlers.
+    async def __aenter__(self) -> 'AsyncServiceBrowser':
+        return self
 
-        At this point the cache will have the new records.
-
-        This method will be run in the event loop.
-        """
-        for pending in self._pending_handlers.items():
-            self._fire_service_state_changed_event(pending)
-        self._pending_handlers.clear()
+    async def __aexit__(
+        self,
+        exc_type: Optional[Type[BaseException]],
+        exc_val: Optional[BaseException],
+        exc_tb: Optional[TracebackType],
+    ) -> Optional[bool]:
+        await self.async_cancel()
+        return None
 
 
 class AsyncZeroconfServiceTypes(ZeroconfServiceTypes):
@@ -180,6 +177,7 @@ class AsyncZeroconf:
         ttl: Optional[int] = None,
         allow_name_change: bool = False,
         cooperating_responders: bool = False,
+        strict: bool = True,
     ) -> Awaitable:
         """Registers service information to the network with a default TTL.
         Zeroconf will then respond to requests for information for that
@@ -192,7 +190,7 @@ class AsyncZeroconf:
         and therefore can be awaited if necessary.
         """
         return await self.zeroconf.async_register_service(
-            info, ttl, allow_name_change, cooperating_responders
+            info, ttl, allow_name_change, cooperating_responders, strict
         )
 
     async def async_unregister_all_services(self) -> None:
@@ -237,11 +235,14 @@ class AsyncZeroconf:
     ) -> Optional[AsyncServiceInfo]:
         """Returns network's service information for a particular
         name and type, or None if no service matches by the timeout,
-        which defaults to 3 seconds."""
-        info = AsyncServiceInfo(type_, name)
-        if await info.async_request(self.zeroconf, timeout, question_type):
-            return info
-        return None
+        which defaults to 3 seconds.
+
+        :param type_: fully qualified service type name
+        :param name: the name of the service
+        :param timeout: milliseconds to wait for a response
+        :param question_type: The type of questions to ask (DNSQuestionType.QM or DNSQuestionType.QU)
+        """
+        return await self.zeroconf.async_get_service_info(type_, name, timeout, question_type)
 
     async def async_add_service_listener(self, type_: str, listener: ServiceListener) -> None:
         """Adds a listener for a particular service type.  This object
